# -*- coding: utf-8 -*-
import gazelib
from os import path
from gazelib.conversion import utils


def convert(gazedata_file_path, experiment_config_file_path,
            participant_number, trial_config_id, was_calibrated):
    '''
    Parameters:
        gazedata_file_path: string
        experiment_config_file_path: string
        participant_number: string
        trial_config_id: e.g. 'SRT1'
        was_calibrated: boolean

    Return:
        CommonV1 object

    '''

    # Read the files
    gdfp = gazedata_file_path  # short alias
    ecfp = experiment_config_file_path
    gd = gazelib.io.load_csv_as_dictlist(gdfp)
    ec = gazelib.io.load_json(ecfp)

    # Convert configuration to object
    expconf = utils.ExperimentConfiguration(ec)
    trialconf = expconf.get_trial_configuration(trial_config_id)

    # Return with a gazelib-common-v1 Object.
    c = gazelib.containers.CommonV1()

    # Build environment constants
    participant_number = str(participant_number)
    c.add_environment('gazelib/gaze/head_id', participant_number)

    trial_config_id = str(trial_config_id)
    c.add_environment('icl/gaze/trial_configuration_id', trial_config_id)

    source_files = [gazedata_file_path, experiment_config_file_path]
    source_files = list(map(path.basename, source_files))
    c.add_environment('gazelib/general/source_files', source_files)

    assert isinstance(was_calibrated, bool)
    c.add_environment('icl/gaze/tracker_successfully_calibrated',
                      was_calibrated)

    c.add_environment('gazelib/gaze/eyetracker/manufacturer', 'Tobii')
    c.add_environment('gazelib/gaze/eyetracker/model', 'TX300')

    c.add_environment('gazelib/gaze/tracked_display_size', {
        "physical_mm": {
            "width": 510.0,
            "height": 288.0
        },
        "resolution_px": {
            "width": 1024,
            "height": 768
        }
    })

    # Build timeline

    def get_microseconds(tet_time):
        '''Takes in time in microseconds'''
        # tet_time might be string, hence int()
        return int(tet_time)

    global_time = get_microseconds(gd[0]['TETTime'])
    # Reference point in microseconds
    global_time_micros = global_time

    def convert_to_relative(tet_time):
        '''Convert to relative times, ints in microseconds'''
        return int(tet_time) - global_time_micros

    def get_relative_time(gazepoint):
        '''Convert gazepoint's TETTime to CommonV1 compliant relative time'''
        return convert_to_relative(gazepoint['TETTime'])

    timeline = list(map(get_relative_time, gd))

    c.set_time_reference(global_time)
    c.add_timeline('eyetracker', timeline)

    # Build streams

    def tobii_validity_to_confidency(validity):
        '''Take in integer validity, return float confidence'''
        validity = int(validity)
        if validity == 0:
            return 1.0
        if validity == 1:
            return 0.8
        if validity == 2:
            return 0.5
        if validity == 3:
            return 0.1
        if validity == 4:
            return 0.0
        raise Exception('Invalid Tobii validity: ' + str(validity))

    left_confidency = []
    right_confidency = []

    left_eye_x_relative_stream = []
    left_eye_y_relative_stream = []
    left_eye_pupil_mm_stream = []
    right_eye_x_relative_stream = []
    right_eye_y_relative_stream = []
    right_eye_pupil_mm_stream = []

    def convert_x(x):
        '''X's already in relative units but may be strings'''
        x = float(x)
        if x == -1.0:
            return None
        return x

    def convert_y(y):
        '''Y's already in relative units but may be strings'''
        y = float(y)
        if y == -1.0:
            return None
        return y

    def convert_pupil(pupil_diam):
        d = float(pupil_diam)
        if d == -1.0:
            return None
        return d

    for gazepoint in gd:
        l_validity = gazepoint['ValidityLeftEye']
        r_validity = gazepoint['ValidityRightEye']
        l_confidency = tobii_validity_to_confidency(l_validity)
        r_confidency = tobii_validity_to_confidency(r_validity)
        left_confidency.append(l_confidency)
        right_confidency.append(r_confidency)

        left_eye_x_relative = convert_x(gazepoint['XGazePosLeftEye'])
        left_eye_y_relative = convert_y(gazepoint['YGazePosLeftEye'])
        left_eye_pupil_mm = convert_pupil(gazepoint['LeftEyePupilDiameter'])
        right_eye_x_relative = convert_x(gazepoint['XGazePosRightEye'])
        right_eye_y_relative = convert_y(gazepoint['YGazePosRightEye'])
        right_eye_pupil_mm = convert_pupil(gazepoint['RightEyePupilDiameter'])

        left_eye_x_relative_stream.append(left_eye_x_relative)
        left_eye_y_relative_stream.append(left_eye_y_relative)
        left_eye_pupil_mm_stream.append(left_eye_pupil_mm)
        right_eye_x_relative_stream.append(right_eye_x_relative)
        right_eye_y_relative_stream.append(right_eye_y_relative)
        right_eye_pupil_mm_stream.append(right_eye_pupil_mm)

    c.add_stream('gazelib/gaze/left_eye_x_relative',
                 'eyetracker', left_eye_x_relative_stream)
    c.add_stream('gazelib/gaze/left_eye_y_relative',
                 'eyetracker', left_eye_y_relative_stream)
    c.add_stream('gazelib/gaze/left_eye_pupil_mm',
                 'eyetracker', left_eye_pupil_mm_stream)
    c.add_stream('gazelib/gaze/right_eye_x_relative',
                 'eyetracker', right_eye_x_relative_stream)
    c.add_stream('gazelib/gaze/right_eye_y_relative',
                 'eyetracker', right_eye_y_relative_stream)
    c.add_stream('gazelib/gaze/right_eye_pupil_mm',
                 'eyetracker', right_eye_pupil_mm_stream)

    # Build events

    #######
    # ICL experiment period tags
    #######
    def value_converter(r):
        tag = r['Tag']
        # Skip nones
        if tag == '':
            raise ValueError()
        if tag == 'Wait':
            return 'icl/experiment/reaction/period/wait'
        if tag == 'AG':
            return 'icl/experiment/reaction/period/attention-grabber'
        if tag == 'FP':
            return 'icl/experiment/reaction/period/pretarget'
        if tag == 'Target':
            return 'icl/experiment/reaction/period/target'
        raise

    def time_converter(r):
        return get_relative_time(r)

    tag_ranges = utils.split_to_ranges_at_change_in_value(gd,
                                                          value_converter,
                                                          time_converter)
    for tag in tag_ranges:
        c.add_event([tag['value']], tag['start'], tag['end'])

    #########
    # Trials
    ##########
    def value_converter(r):
        if r['Trialnumber'] == '':
            raise ValueError()
        return int(r['Trialnumber'])  # Raises value error if not int

    def time_converter(r):
        return get_relative_time(r)

    ranges = utils.split_to_ranges_at_change_in_value(gd,
                                                      value_converter,
                                                      time_converter)
    for r in ranges:
        extra = {
            'icl/experiment/reaction/trial/sequence_number': r['value']
        }
        c.add_event(['icl/experiment/reaction/trial'],
                    r['start'], r['end'], extra=extra)

    ##################
    # Image stimuli i.e. UserDefined_1 & Aoi
    ##################
    def value_converter(r):
        # Split when at least one of the two changes
        # Interpret empty values as invalid.
        if r['UserDefined_1'] == '' or r['Aoi'] == '':
            raise ValueError()
        return r['UserDefined_1'] + '-' + r['Aoi']

    ranges = utils.split_to_ranges_at_change_in_value(gd, value_converter,
                                                      time_converter)
    for r in ranges:
        original_row = r['first']
        image_index = int(original_row['UserDefined_1'])
        aoi_index = int(original_row['Aoi'])
        extra = {
            'original_area_of_interest_index': aoi_index,
            'original_image_index': image_index,
            'icl/stimulus/image/filename':
                trialconf.get_image_name(image_index),
            'icl/stimulus/image/rectangle':
                trialconf.get_aoi_rectangle(aoi_index)
        }
        c.add_event(['icl/stimulus', 'icl/stimulus/image'],
                    r['start'], r['end'], extra=extra)

    return c



'''
igazelib: A low level library for analyzing gaze-data files provided usually
by eyetracker-devices.

Gaze data is expected to be in JSON-format such as list of datapoints with
similar dict storing the properties for each point.

Example:
  [{xcoordinate:0.4, ...}, {xcoordinate:0.5, ...}, ..., {xcoordinate:-1, ...}]

List elements are called "gazepoints", and dict keys as "keys".

gazelib-library is designed to be used with a script file that calls
the library functions to perform different analysis steps.

Created by researchers in Infant Cognition Lab,
University of Tampere, Finland
'''

indent = "  "


def median(lst):
    '''
    Return median of the list of numbers.
    '''
    # http://stackoverflow.com/a/24101534/638546
    sortedLst = sorted(lst)
    lstLen = len(lst)
    index = (lstLen - 1) // 2

    if (lstLen % 2):
        return sortedLst[index]
    else:
        return (sortedLst[index] + sortedLst[index + 1]) / 2.0


def mean(lst):
    '''
    Return mean of the list of numbers.
    '''
    # http://stackoverflow.com/a/7716358/638546
    return float(sum(lst)) / len(lst) if len(lst) > 0 else float('nan')


def first_gazepoints_by_time(data, time_key, timeunits, silent=True):
    '''
    Clip first rows from DATA before milliseconds count of time has passed
    '''

    if not silent:
        print("Picking first " + str(timeunits) +
              " timeunit gazepoints from data...")
        print(indent + "List contains " + str(len(data)) +
              " gazepoints before operation.")

    tk = time_key  # short alias
    start = int(data[0][tk])
    new_data = [gp.copy() for gp in data if timeunits > int(gp[tk]) - start]

    if not silent:
        print(indent + "List contains " + str(len(new_data)) +
              " gazepoints after operation.")

    return new_data


def first_gazepoints(data, gpcount, silent=True):
    '''
    Clip first rows from DATA before milliseconds count of time has passed
    '''

    if not silent:
        print("Picking first " + str(gpcount) + " gazepoints from data...")
        print(indent + "List contains " + str(len(data)) +
              " gazepoints before operation.")

    new_data = [gp.copy() for index, gp in enumerate(data) if index < gpcount]

    if not silent:
        print(indent + "List contains " + str(len(new_data)) +
              " gazepoints after operation.")

    return new_data


def gazepoints_after_time(data, time_key, timeunits, silent=True):
    '''
    Clip rows from DATA after [timeunits] count of time has passed
    '''

    if not silent:
        print("Picking gazepoints after " + str(timeunits) +
              " timeunits from data using TETTime...")
        print(indent + "List contains " + str(len(data)) +
              " gazepoints before operation.")

    tk = time_key  # short alias

    # find the time of the first datapoint of the list
    start = int(get_value(data, 0, tk))

    # generate a new list of datapoints
    new_data = [gp.copy() for gp in data if timeunits <= int(gp[tk]) - start]

    if not silent:
        print(indent + "List contains " + str(len(new_data)) +
              " gazepoints after operation.")

    return new_data


def gazepoints_containing_value(data, key, value_list, silent=True):
    '''
    return the rows that contain certain value
    '''

    if not silent:
        print("Picking gazepoints with values " + str(value_list) +
              " assosiated with key " + str(key))
        print(indent + "List contains " + str(len(data)) +
              " gazepoints before operation.")

    # check if input is really a list
    assert isinstance(value_list, list)

    # find gazepoints which contain the value
    gazepoints_found = [gp.copy() for gp in data if gp[key] in value_list]

    if not silent:
        print(indent + "Datamatrix contains " + str(len(gazepoints_found)) +
              " gazepoints.")

    return gazepoints_found


def gazepoints_not_containing_value(data, key, value_list, silent=True):
    '''
    return the rows that DO NOT contain certain value
    '''

    if not silent:
        print("Picking gazepoints without values " + str(value_list) +
              " assosiated with key " + str(key))
        print(indent + "List contains " + str(len(data)) + " gazepoints.")

    # check if input is really a list
    assert isinstance(value_list, list)

    # find rows which do not contain the value
    rows_found = [row for row in data if row[key] not in value_list]

    if not silent:
        print(indent + "List contains " + str(len(rows_found)) +
              " gazepoints.")

    return rows_found


def split_at_change_in_value(data, key, silent=True):
    '''
    Split one list to multiple lists. New list is started each time when
    value in the column changes from previous one. Omitted on the case of
    first element.
    '''

    # return list of lists
    if not silent:
        print("Splitting data when change in value for key: " + str(key))

    list_of_new_datas = []

    new_data = []
    previous_item = None

    # loop throught all gazepoints
    for gp in data:

        # if there is gazepoints before
        if previous_item is not None:

            # if previous item does not match the one we are processing
            if not previous_item == gp[key]:
                list_of_new_datas.append(new_data)
                new_data = []

        new_data.append(gp.copy())
        previous_item = gp[key]

    # append the last remaining clip when all looped
    list_of_new_datas.append(new_data)

    if not silent:
        print(indent + "Returning " + str(len(list_of_new_datas)) +
              " gazepoint lists.")

    return list_of_new_datas


def get_value(data, gazepoint, key):
    '''
    Returns a value from specific datapoint with specific key.
    '''

    return data[gazepoint][key]


def replace_value(data, key, value_to_replace, value, silent=True):
    '''
    replaces all values value_to_replace on key with value.
    '''

    if not silent:
        print("Replacing values for key:" + str(key) + " with value " +
              str(value_to_replace))

    vtr = value_to_replace
    new_values = [value if gp[key] == vtr else gp[key] for gp in data]
    new_data = add_key(data, key, new_values)

    if not silent:
        print(indent + "Done.")

    return new_data


def border_violation(data, aoi, xkey, ykey, valkey,
                     accepted_validities, silent=True):
    '''
    Return true if during-non valid perioid gaze has crossed aoi-border.
    '''

    if not silent:
        print("Calculating if a gaze moved over aoi border " +
              "during invalid data...")

    gaze_in_last_good = True
    gaze_okay_before = True

    for index, row in enumerate(data):
        # if accepted_validities.__contains__(row[valcol]):
        if row[valkey] in accepted_validities:
            # gaze okay
            gaze_okay = True
        else:
            # gaze not okay
            gaze_okay = False

        gaze_in = inside_aoi(aoi, row[xkey], row[ykey])

        if index > 0:
            if gaze_okay:
                if not gaze_okay_before and gaze_in != gaze_in_last_good:
                    if not silent:
                        print(indent + "Border violation detected.")
                    return True

                gaze_in_last_good = gaze_in

        gaze_okay_before = gaze_okay

    if not silent:
        print(indent + "No border violation detected.")

    return False


def inside_aoi(aoi, x, y):
    '''
    If coordinates are inside aoi, return true, otherwise false.
    Aoi is expected to be a dict with keys "x1", "x2", "y1", "y2".
    Aoi represents an rectangle with top-left corner at (x1, y1) and
    bottom-right corner at (x2, y2).
    '''

    if aoi["x1"] < x and x < aoi["x2"] and aoi["y1"] < y and y < aoi["y2"]:
        return True
    else:
        return False


def combine_coordinates(data, accepted_validities, rxkey, rykey, rvalkey,
                        lxkey, lykey, lvalkey, silent=True):
    '''
    Combine two coordinate-columns with third validity-column to one column.
    '''

    if not silent:
        print("Combining two columns...")

    x = []
    y = []
    val = []

    # loop all rows and collect x, y coordinates +
    # minimum validity value (assumed to be best)
    for row in data:
        x.append(mean_of_valid_values([row[rxkey], row[lxkey]],
                                      [row[rvalkey], row[lvalkey]],
                                      accepted_validities))
        y.append(mean_of_valid_values([row[rykey], row[lykey]],
                                      [row[rvalkey], row[lvalkey]],
                                      accepted_validities))
        val.append(min(int(row[rvalkey]), int(row[lvalkey])))

    if not silent:
        print(indent + "Done.")

    return x, y, val


def add_key(data, key, new_values):
    '''
    Adds a key to the datapoint-list. New_values must match in length to
    the column with keys.
    '''

    new_data = []
    for index, gp in enumerate(data):
        new_gp = gp.copy()       # use copy-method to not only affect pointer
        new_gp[key] = new_values[index]
        new_data.append(new_gp)

    return new_data


def get_key(data, key):
    '''
    Returns a list of values, from single key in the data-parameter.
    '''

    column = []
    for row in data:
        column.append(row[key])
    return column


def median_filter_data(data, winlen, key):
    '''
    Performs median filtering to the datapoints in the specified column on
    DATA-structure. Further information see help medianFilter.
    Column specifies the column to filter the data with.
    '''

    new_data = add_key(data, key, median_filter(get_key(data, key), winlen))

    return new_data


def median_filter(datapoints, winlen, silent=True):
    '''
    Performs median filtering to the datapoints with window-length winlen.
    Winlen must be an odd integer
    (window: sample-(winlen-1)/2..sample..sample+(winlen-1)/2.
    Endings of the sample are truncated by the first/last sample to achieve
    filtered trace of same length than the original.
    Here datapoints must contain numbers, otherwise an error is presented.
    '''

    datapoints = list(map(float, datapoints))

    if not silent:
        print("Performing median filtering with window-length " +
              str(winlen) + " for " + str(len(datapoints)) + " datapoints...")

    # calculate padding length
    padlen = (winlen - 1) // 2

    # form padding (first and last number repeated at the beginning and end)
    pad_start = padlen * [datapoints[0]]
    pad_end = padlen * [datapoints[-1]]

    datapoints_pad = pad_start + datapoints + pad_end

    # for each datapoint
    filtered_datapoints = []
    for i in range(0, len(datapoints)):
        wind = datapoints_pad[i:i+2*padlen+1]
        filtered_datapoints.append(median(wind))

    if not silent:
        print(indent + "Done.")

    return filtered_datapoints


def interpolate_using_last_good_value(data, key, valkey, accepted_validities,
                                      silent=True):
    '''
    Interpolates values with key "key" in DATA-matrix by replacing the bad
    value with last good value before bad values (if there is at least one good
    value, otherwise, do nothing). Validitycolumn contains the validity
    markings for each datapoint and good validities are defined by the accepted
    validities-parameter. If the beginning of a trail is "bad", use the first
    appearing good value to interpolate that.
    '''

    if not silent:
        print("Interpolating values + " + key +
              ": using last good (or first good) value...")

    # find the first non-bad, if any

    first_valid = -1
    for gpnum, gp in enumerate(data):

        if gp[valkey] in accepted_validities and first_valid == -1:
            first_valid = gpnum

    # there was at least one good value
    if first_valid == -1:
        # if not, return data as it was
        if not silent:
            print(indent + "Done. No good data available")
        return data

    new_data = []

    last_non_bad = get_value(data, first_valid, key)
    for gp in data:
        if gp[valkey] in accepted_validities:
            # valid data
            last_non_bad = gp[key]
        else:
            gp[key] = last_non_bad

        new_data.append(gp)

    if not silent:
        print(indent + "Done.")

    return new_data


def gaze_inside_aoi(data, xcol, ycol, aoi, firstorlast, silent=True):
    '''
    Finds either the first row when gaze enters aoi or last.
    If gaze does not enter aoi, return -1
    '''

    if not silent:
        print("Calculating when gaze inside aoi: " + str(aoi) +
              " " + firstorlast + "time...")

    last_in = -1
    for rownumber, row in enumerate(data):
        if inside_aoi(aoi, row[xcol], row[ycol]):
            if firstorlast == "first":
                if not silent:
                    print(indent + "Done.")
                return rownumber

            last_in = rownumber

    if not silent:
        print(indent + "Done.")

    return last_in


def gaze_inside_aoi_percentage(data, xcol, ycol, aoi, silent=True):
    '''
    Calculate the percentage of gaze inside aoi borders on given data.
    '''

    if not silent:
        print("Calculating when the portion of gaze inside aoi: " + str(aoi))

    rowcount = len(data)

    if rowcount == 0:
        return -1

    gaze_inside = 0
    for row in data:
        if inside_aoi(aoi, row[xcol], row[ycol]):
            gaze_inside = gaze_inside + 1

    if not silent:
        print(indent + "Done.")

    return gaze_inside / float(rowcount)


def longest_non_valid_streak(data, valkey, timekey, accepted_validities,
                             silent=True):
    '''Longest streak of non valid values'''

    if not silent:
        print("Calculating longest non-valid streak...")

    streak_started = None
    longest_streak = 0
    streak_on = False

    for gp in data:

        # check if ongoing streak and put to longest if is
        if streak_on:
            # Assert streak_started is set if streak_on
            streaklen = float(gp[timekey]) - float(streak_started)
            if streaklen > longest_streak:
                longest_streak = streaklen

        # test if this datapoint valid and start streak or end it
        valid = gp[valkey] in accepted_validities
        if valid:
            streak_on = False
        else:
            if not streak_on:
                streak_started = gp[timekey]
            streak_on = True

    if not silent:
        print(indent + "Done.")

    return longest_streak


def valid_gaze_percentage(data, valkey, accepted_validities):
    '''Calculates the percentage of valid gaze in data.'''

    if len(data) == 0:
        return -1

    valid = 0

    for gp in data:
        if gp[valkey] in accepted_validities:
            valid = valid + 1

    return float(valid) / float(len(data))


def duration(data, timekey):
    '''Returns the length of the data in time units.'''

    if len(data) == 0:
        return -1
    elif len(data) <= 1:
        return 0.0

    end_time = float(get_value(data, -1, timekey))
    start_time = float(get_value(data, 0, timekey))
    return end_time - start_time


def SRT_index(rtimes, max_rt, min_rt):
    '''calculate SRT index (Leppanen et al.)'''

    SRTs = []

    for rtime in rtimes:
        SRTs.append(1 - (max_rt-rtime) / (max_rt-min_rt))

    return mean(SRTs)


def group(data, group_key, value_key):
    '''
    Groups a list of datapoint-dicts so that one of the keys is used as
    "grouping key" - according to this the values specified by "value key" are
    sorted to the dict containing lists.
    '''

    datas_by_group = {}
    for datapoint in data:

        group = datapoint[group_key]
        value = datapoint[value_key]

        # if already entry of this type of group_key in the dict
        if group in datas_by_group:
            datas_by_group[group].append(value)
        else:
            datas_by_group[group] = [value]

    return datas_by_group


def group_lists(datas, group_key):
    '''
    Groups data by 0'th value in group_column and places each subset to a group
    defined by group_column.
    parameter: list of datas (data:list of rows), grouping key
    value in datapoint[0][group_key] expected to exist
    '''

    datas_by_group = {}
    for data in datas:
        group = get_value(data, 0, group_key)

        # if already entry of this key in the
        if group in datas_by_group:
            datas_by_group[group] = datas_by_group[group] + data
        else:
            datas_by_group[group] = data

    return datas_by_group


def mean_of_valid_values(values, validities, accepted_validities):
    '''
    Returns mean of good validity tagged values. If none, returns -1.
    values: list, validities:list
    '''

    goodvalues = []

    for index, i in enumerate(values):
        if validities[index] in accepted_validities:
            goodvalues.append(float(i))

    if len(goodvalues) == 0:
            return -1
    else:
        return mean(goodvalues)



# -*- coding: utf-8 -*-


class ConversionException(Exception):
    pass


def estimate_sampling_interval(times):
    '''
    Return mean interval in the given list of times.
    Return None if times is empty or contains only one element.
    '''

    n = 0
    dt_sum = 0.0
    prev_t = 0.0
    first = True
    for t in times:
        if first:
            first = False
        else:
            dt_sum += t - prev_t
        prev_t = t
        n += 1

    if n < 2:
        return None

    return dt_sum / (n - 1)


def split_to_ranges_at_change_in_value(gd, value_converter, time_converter):
    '''
    Similar to gazelib.legacy.igazelib.split_at_change_in_value but
    returns also starting and ending times. The start time is equal to
    the end time of the previous range. Invalid values do not cause splits.

    Parameters:
        gd:
            gazedata as list of dicts
        value_converter:
            A function that converts from raw row to the value
            used for comparison and also in yielded dicts.
            Must raise ValueError or TypeError if value cannot be converted.
            On ValueError or TypeError the row will be skipped.
        time_converter:
            A function that converts from raw row to times used
            in ranges.
            Must return integer.
            Must raise ValueError or TypeError if time cannot be converted.
            On ValueError or TypeError the row will be skipped.

    Yields dicts::

        {
            'start': <integer, start time of range, inclusive>,
            'end': <integer, end time of range, exclusive>,
            'value': <the value of the range>,
            'first': <first point in range>
        }

    '''

    # Estimate sample interval (i.e. 1 / sampling rate)
    times = map(time_converter, gd)
    sample_interval = int(round(estimate_sampling_interval(times)))

    cur_time = None
    previous_value = None
    event_to_yield = None
    for index, gazepoint in enumerate(gd):
        try:
            cur_value = value_converter(gazepoint)
            cur_time = time_converter(gazepoint)
            # cur_time must be last because otherwise it would not always
            # point to the time of last valid row. If last rows are
            # invalid, the last event should still have correct time.
        except (ValueError, TypeError):
            # Skip invalid but remember the previous value.
            # This way we do not split at invalid values.
            continue

        # Iterate until value differs from the previous
        if cur_value == previous_value:
            continue

        # Assert: different valid value observed

        # If first row, do not store dict but start the first dict.
        if event_to_yield is None:
            event_to_yield = {
                'start': cur_time,
                'value': cur_value,
                'first': gazepoint
            }
        else:
            # Range dict finished
            event_to_yield['end'] = cur_time
            yield event_to_yield

            # Start new one
            event_to_yield = {
                'start': cur_time,
                'value': cur_value,
                'first': gazepoint
            }

        previous_value = cur_value

    # Handle the last event
    if event_to_yield is not None:
        event_to_yield['end'] = cur_time + sample_interval
        yield event_to_yield


class ExperimentConfiguration(object):

    def __init__(self, raw_dict):
        self.raw = raw_dict

    def get_trial_configuration(self, trial_config_id):
        '''
        Available trial configuration ids are:
            'calibration_movie', 'calibration_stimuli',
            'SRT1', 'SRT2', 'SRT3', 'SRT4', 'SRT5'
        '''
        raw = next(cf for cf in self.raw if cf['name'] == trial_config_id)
        return TrialConfiguration(raw)


class TrialConfiguration(object):

    def __init__(self, raw_dict):
        self.raw = raw_dict

    def get_image_name(self, index):
        '''
        Retrieves an image name from 'images' list at given index.
        '''
        return self.raw['images'][index]

    def get_aoi_rectangle(self, aoi_index):
        '''
        Returns AoI in gazelib/geom/rectangle format.
        '''
        aoi = self.raw['aois'][aoi_index]
        # Convert to gazelib/geom/rectangle
        rect = [aoi[0], aoi[2], aoi[1], aoi[3]]
        return rect



# -*- coding: utf-8 -*-
'''
Input-Output functions that help to read and write gaze data.

'''
import json
import csv


def load_json(filename):
    '''
    Load json-type file and return its contents as python object.

    Raise IOError if loading failed.
    Raise ValueError if JSON decoding failed.

    Note: IOError might vary between environments.
    See: http://stackoverflow.com/q/15032108/638546
    '''
    # It seems that with open... hides exception.
    data_file = open(filename, 'r')  # Can raise IOError
    try:
        data = json.load(data_file)
    finally:
        # Regardless whether JSON load fails, close the file.
        data_file.close()
    return data


def write_json(filename, data, human_readable=False):
    '''
    Dump data to a given JSON file. File is created if it does not exist.

    Raise TypeError if data is not serializable.
    '''
    with open(filename, 'w') as jsonfile:
        if not human_readable:
            json.dump(data, jsonfile)
        else:
            # ensure_ascii=False
            #   'True' would turn non-ascii characters such
            #   as ä and ö to escaped unicode sequences: \uXXXX,
            #   turning text hard to read.
            # separators
            #   Without setting separators explicitly,
            #   Python 2.7 and Python 3.5 would produce different output.
            #   See https://docs.python.org/3/library/json.html#json.dump
            json.dump(data, jsonfile, sort_keys=True, indent=4,
                      ensure_ascii=False, separators=(',', ': '))


def write_fancy_json(filename, data):
    '''
    Write JSON so that it is readable by humans (new lines, indent..).

    Raise (see write_json)
    '''
    return write_json(filename, data, human_readable=True)


def load_csv_as_dictlist(filepath, delimiter='\t'):
    '''
    Load a file in csv (common in .gazedata) and return data as
    list of dicts. Each dict contains the column names as keys and
    cell values as dict values.

    Raise IOError if filepath is invalid
    Raise ValueError if file is not CSV formatted.
    '''

    ifile = open(filepath, 'r')  # Can raise IOError

    try:
        reader = csv.reader(ifile, delimiter=delimiter)

        rownum = 0
        rows = []
        for row in reader:
            items = {}
            if rownum == 0:
                # headers
                headers = row
            else:
                for inum, i in enumerate(headers):
                    items[i] = row[inum]
                rows.append(items)
            rownum += 1
    finally:
        ifile.close()

    if rows == []:
        raise ValueError('CSV file is not correctly formatted.')

    return rows


def write_dictlist_as_csv(target_filename, dictlist, headers=None,
                          delimiter='\t'):
    '''
    Write given list of dicts to a file as CSV. Each dict is required
    to have key for each column name. The column names are read from the
    first dict.

    Parameters:
        target_filename:
            A path to target file
        dictlist:
            A list of dicts. Each dict is required to have same set of keys.
        headers:
            Optional list of keys. Defines the order of columns.
        delimiter:
            Delimiter character.
    '''
    # Normalize to iterable
    dictlist = iter(dictlist)
    first = next(dictlist)
    # Headers
    if headers is None:
        fieldnames = first.keys()
    else:
        fieldnames = headers
    # Write
    with open(target_filename, 'w') as csvfile:
        writer = csv.DictWriter(csvfile,
                                fieldnames=fieldnames,
                                lineterminator='\n',
                                delimiter=delimiter)
        writer.writeheader()
        writer.writerow(first)
        for d in dictlist:
            writer.writerow(d)



# -*- coding: utf-8 -*-
'''
Classes that store the gaze data and can be fed to analysis functions.
'''
from .validation import is_list_of_strings, is_integer, is_string
# from .settings import min_event_slice_overlap_seconds as min_overlap
from .statistics import arithmetic_mean, deltas
from .io import load_json, write_json, write_fancy_json, write_dictlist_as_csv
from time import time as get_current_posix_time
from deepdiff import DeepDiff
from bisect import bisect_left  # binary tree search tool
from jsonschema import validate as validate_jsonschema


def get_current_time_reference():
    '''Get current POSIX time in microseconds'''
    return int(get_current_posix_time() * 10**6)


class CommonV1(object):

    class InvalidRangeException(Exception):
        '''
        Raised if invalid range index pair were given. E.g. start > end
        '''
        pass

    class MissingEnvironmentException(Exception):
        '''Raised if a given environment name cannot be found.'''
        pass

    class MissingTimelineException(Exception):
        '''Raised if a given timeline name cannot be found.'''
        pass

    class MissingStreamException(Exception):
        '''Raised if a given stream name cannot be found.'''
        pass

    class MissingTagException(Exception):
        '''Raised if a given event tag cannot be found'''

    class InvalidTimelineException(Exception):
        '''Raised if a timeline does not fit the CommonV1 specification.'''
        pass

    class InvalidStreamException(Exception):
        '''Raised if a stream does not fit the CommonV1 specification.'''
        pass

    class InvalidEventException(Exception):
        '''Raised if an event does not fit the CommonV1 specification.'''
        pass

    class InsufficientDataException(Exception):
        '''Raised if container does not provide required data.
        For example if some streams are missing.'''
        pass

    class InvalidTimeException(Exception):
        '''Raised if a time is not an integer.'''
        pass

    class EmptyContainerException(Exception):
        '''
        Raised if there is not enough content for a CommonV1 method.
        For example get_relative_start_time cannot return anything
        if there is no timelines or they are empty.
        '''
        pass

    # JSON Schema to validate raw input
    # For grammar,
    # see http://json-schema.org/latest/json-schema-validation.html
    SCHEMA = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'title': 'gazelib/common/v1',
        'type': 'object',
        'properties': {
            'schema': {
                'type': 'string',
                'pattern': 'gazelib/common/v1'
            },
            'time_reference': {
                'type': 'integer'
            },
            'environment': {
                'type': 'object',
                'patternProperties': {
                    '.+': {}
                }
            },
            'timelines': {
                'type': 'object',
                'patternProperties': {
                    '.+': {
                        'type': 'array',
                        'items': {
                            'type': 'integer'
                        }
                    }
                }
            },
            'streams': {
                'type': 'object',
                'patternProperties': {
                    '.+': {
                        'type': 'object',
                        'properties': {
                            'timeline': {
                                'type': 'string'
                            },
                            'values': {
                                'type': 'array'
                            },
                            'derived': {
                                'type': 'string'
                            },
                            'confidence': {
                                'type': 'array',
                                'items': {
                                    'type': 'number',
                                    'maximum': 1.0,
                                    'minimum': 0.0
                                }
                            }
                        },
                        'required': ['timeline', 'values']
                    }
                }
            },
            'events': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'tags': {
                            'type': 'array',
                            'items': {
                                'type': 'string'
                            }
                        },
                        'range': {
                            'type': 'array',
                            'items': {
                                'type': 'number'
                            }
                        },
                        'derived': {
                            'type': 'string'
                        },
                        'extra': {}
                    },
                    'required': ['tags', 'range']
                }
            }
        },
        'required': ['schema', 'time_reference', 'environment',
                     'timelines', 'streams', 'events']
    }

    @staticmethod
    def validate(raw_common):
        '''
        Raises ValidationError if raw_common is not valid gazelib/common/v1
        '''
        validate_jsonschema(raw_common, CommonV1.SCHEMA)

    def __init__(self, raw_common_or_filepath=None):
        '''
        Parameters:
            raw_common_or_filepath:
                Optional. A dict formatted in gazelib/common/v1
                or a string path to a source file.
                Supported source file types:
                - JSON

        Raises:
            ValidationError
        '''

        r = raw_common_or_filepath  # alias

        if r is None:
            # Construct with empty.
            now = get_current_time_reference()
            r = {
                'schema': 'gazelib/common/v1',
                'time_reference': now,
                'environment': {},
                'timelines': {},
                'streams': {},
                'events': []
            }
        elif isinstance(r, str):
            # Load file
            r = load_json(r)
            CommonV1.validate(r)
        else:
            CommonV1.validate(r)

        self.raw = r

    def __eq__(self, other):
        '''Override '==' operator with deep difference check.'''
        return DeepDiff(self.raw, other.raw) == {}

    # Assertions

    def assert_has_environments(self, env_names):
        '''
        Raises CommonV1.InsufficientDataException if the any of the
        given environments is missing from the container. This is used
        in analysis scripts that require certain environments.
        '''
        if not self.has_environments(env_names):
            required = ', '.join(env_names)
            available = ', '.join(self.get_environment_names())
            msg = 'Environments ' + required + ' are needed but only ' + \
                available + ' are available.'
            raise CommonV1.InsufficientDataException(msg)

    def assert_has_streams(self, stream_names):
        '''
        Raises CommonV1.InsufficientDataException if the any of the
        given streams is missing from the container. This is used
        in analysis scripts that require certain streams.
        '''
        if not self.has_streams(stream_names):
            required = ', '.join(stream_names)
            available = ', '.join(self.get_stream_names())
            msg = 'Streams ' + required + ' are needed but only ' + \
                available + ' are available.'
            raise CommonV1.InsufficientDataException(msg)

    def assert_range_order(self, start, end, start_name, end_name):
        '''
        Raise InvalidRangeException if start larger or equal than end.
        '''
        # Ensure correct order
        if start is not None and end is not None:
            if start >= end:
                msg = (start_name + ' (' + str(start) + ') must be ' +
                       'smaller than ' + end_name + ' (' + str(end) + ')')
                raise CommonV1.InvalidRangeException(msg)

    # Accessors

    def convert_to_unix_time(self, relative_time_micros):
        '''Return unix time in microseconds (int)'''
        if relative_time_micros is None:
            return None
        tr = self.raw['time_reference']
        return tr + relative_time_micros

    def convert_to_relative_time(self, unix_time_micros):
        '''Return relative time in microseconds (int)'''
        if unix_time_micros is None:
            return None
        gt = self.raw['time_reference']
        return unix_time_micros - gt

    def count_events(self, tag=None):
        '''
        Return integer, number of stored events.
        If the optional tag is given, count only events with the tag.
        '''
        if tag is None:
            return len(self.raw['events'])
        else:
            total = 0
            for ev in self.raw['events']:
                if tag in ev['tags']:
                    total += 1
            return total

    def get_duration(self):
        '''
        Difference in microseconds between the smallest and largest time point.
        '''
        return self.get_relative_end_time() - self.get_relative_start_time()

    def get_environment(self, env_name):
        '''
        Return value of the environment.
        '''
        try:
            return self.raw['environment'][env_name]
        except KeyError:
            str_e = str(env_name)
            msg = 'Environment ' + str_e + ' not found.'
            raise CommonV1.MissingEnvironmentException(msg)

    def get_environment_names(self):
        '''
        DEPRECATED: Use list_environment_names instead
        '''
        return self.list_environment_names()

    def get_event_by_tag(self, tag, index=0):
        '''
        Return event that has the given tag. By default the first occurence
        is returned.

        Raise MissingTagException if no events are found.
        Raise IndexError if tags found but index out of range.
        '''

        # Select event of the index:th tag
        tag_found = False
        i = 0
        for event in self.raw['events']:
            if tag in event['tags']:
                tag_found = True
                if index == i:
                    return event
                i += 1

        # Tag was found but index was too high.
        if tag_found:
            raise IndexError('Index ' + str(index) + ' out of range(' +
                             str(i) + ').')
        else:
            raise CommonV1.MissingTagException('Events with tag ' + str(tag) +
                                               ' not found.')

    def get_relative_start_time(self):
        '''
        Find he smallest time in the container as relative time.
        Can be negative.
        '''

        tls = self.raw['timelines']
        evs = self.raw['events']

        has_timelines = len(tls.keys()) > 0
        has_events = len(evs) > 0
        # Timelines cannot be empty

        if has_timelines:
            # Min of first elements in timelines
            tl_1st = [tl[0] for tl in tls.values()]
            tl_min = min(tl_1st)

        if has_events:
            # Min in range of events
            ev_1st = [ev['range'][0] for ev in evs]
            ev_min = min(ev_1st)

        if has_timelines:
            if has_events:
                return min(tl_min, ev_min)
            else:
                return tl_min
        else:
            if has_events:
                return ev_min
            else:
                raise CommonV1.EmptyContainerException('No time content.')

    def get_relative_end_time(self):
        '''
        Find the largest time in the container as relative time.
        Can be negative.
        '''

        tls = self.raw['timelines']
        evs = self.raw['events']

        has_timelines = len(tls.keys()) > 0
        has_events = len(evs) > 0

        if has_timelines:
            # Max of last elements in timelines
            tl_last = [tl[-1] for tl in tls.values()]
            tl_max = max(tl_last)

        if has_events:
            # Max in range of events
            ev_last = [ev['range'][1] for ev in evs]
            ev_max = max(ev_last)

        if has_timelines:
            if has_events:
                return max(tl_max, ev_max)
            else:
                return tl_max
        else:
            if has_events:
                return ev_max
            else:
                raise CommonV1.EmptyContainerException('No time content.')

    def get_relative_time_by_index(self, timeline_name, index):
        '''
        Return relative time on the timeline at index.
        '''
        tl = self.get_timeline(timeline_name)
        return tl[index]

    def get_stream_values(self, stream_name):
        '''
        Return: the full list of values of the stream.
        Raise: CommonV1.MissingStreamException: if name not found.
        '''
        return self.get_stream(stream_name)['values']

    def get_stream_timeline_name(self, stream_name):
        '''
        Return: the timeline name associated with the stream.
        Raise: CommonV1.MissingStreamException: if name not found.
        '''
        return self.get_stream(stream_name)['timeline']

    def get_stream(self, stream_name):
        '''
        Return: CommonV1 stream dict.
        Raise: CommonV1.MissingStreamException: if name not found.
        '''
        try:
            return self.raw['streams'][stream_name]
        except KeyError:
            str_name = str(stream_name)
            raise CommonV1.MissingStreamException('Stream ' + str_name +
                                                  ' not found.')

    def get_stream_names(self):
        '''
        DEPRECATED: Use list_stream_names instead
        '''
        return self.list_stream_names()

    def get_timeline(self, timeline_name):
        '''
        Return:
            timeline i.e. a list of relative times.

        Raise:
            CommonV1.MissingTimelineException: if name not found.
        '''
        try:
            return self.raw['timelines'][timeline_name]
        except KeyError:
            str_tl = str(timeline_name)
            raise CommonV1.MissingTimelineException('Timeline ' + str_tl +
                                                    ' not found.')

    def get_timeline_mean_interval(self, timeline_name):
        '''
        Return mean interval in microseconds, measured over the timeline.
        '''
        try:
            tl = self.raw['timelines'][timeline_name]
            return arithmetic_mean(deltas(tl))
        except KeyError:
            str_tl = str(timeline_name)
            raise CommonV1.MissingTimelineException('Timeline ' + str_tl +
                                                    ' not found.')

    def get_time_reference(self):
        '''
        Return the time reference as microseconds from unix epoch.
        '''
        return self.raw['time_reference']

    def has_environments(self, env_names):
        '''
        Test if environments are available.

        Parameter:
            env_names, list of strings

        Return:
            True, if all the given environments are available.
            False, otherwise
        '''
        available_envs = self.get_environment_names()
        return all(env in available_envs for env in env_names)

    def has_streams(self, stream_names):
        '''
        Test if streams are available.

        Parameter:
            stream_names, list of strings

        Return:
            True, if all the given streams are available.
            False, otherwise
        '''
        available_streams = self.get_stream_names()
        return all(st in available_streams for st in stream_names)

    def iter_events(self):
        '''
        Iterate over each event. See list_events to get list directly.
        '''
        return iter(self.raw['events'])

    def iter_events_by_tag(self, tag):
        '''
        Yield events that have the tag.
        Raise MissingTagException if no events found.
        '''
        for ev in self.iter_events_by_tags([tag]):
            yield ev

    def iter_events_by_tags(self, tags):
        '''
        Yield events that have at least one of the tags.
        For example, used internally by save_events_as_csv(tags)

        Parameters:
            tags: a list of tags

        If no such events found, iter []
        '''
        num_yields = 0
        stags = set(tags)
        for ev in self.iter_events():
            # Intersection
            sev = set(ev['tags'])
            if len(stags.intersection(sev)) > 0:
                yield ev
                num_yields += 1
        if num_yields == 0:
            return  # Generators allow single return

    def iter_by_tag(self, tag, limit_to=None):
        '''DEPRECATED as too vague. Use iter_slices_by_tag instead.'''
        return self.iter_slices_by_tag(tag, limit_to)

    def iter_slices_by_tag(self, tag, limit_to=None):
        '''
        Slice to multiple portions. E.g. if there is ten events with "trial"
        tag, returns iterable over ten slices, one for each tag.

        Parameters
            tag
                string
            limit_to
                max number of slices to return. By default returns all.

        Return
            iterable of CommonV1 objects

        If no such slices found, iter []
        '''
        for index, event in enumerate(self.iter_events_by_tag(tag)):
            range_start = event['range'][0]
            range_end = event['range'][1]
            yield self.slice_by_relative_time(range_start, range_end)
            if limit_to is not None:
                if index + 2 > limit_to:
                    break

    def list_environment_names(self):
        '''Return list of names of provided environments.'''
        return list(self.raw['environment'].keys())

    def list_events(self):
        '''Return full list of events'''
        return self.raw['events']

    def list_stream_names(self):
        '''Return list of names of provided streams.'''
        return list(self.raw['streams'].keys())

    def list_tags(self):
        '''Return list of unique tags in the events. Order is not defined.'''

        tags = {}  # Looping lists is O(n). Searching a key is O(log(n))

        for ev in self.iter_events():
            for tag in ev['tags']:
                if tag not in tags:
                    tags[tag] = True

        return list(tags.keys())

    def slice_by_relative_time(self, rel_start_time, rel_end_time=None):
        '''
        Return new CommonV1 object with data only in the time range.
        Will slice events that are partially outside.
        Will remove empty streams and events that are completely outside.
        Will remove empty timelines
        Does not update time_reference because easier implementation
        and more efficient execution.
        '''

        # Ensure given range is in correct order
        self.assert_range_order(rel_start_time, rel_end_time,
                                'rel_start_time', 'rel_end_time')

        # The new copy of content. Insert sliced values to it.
        # time reference, schema, and environment stay the same.
        slice_raw = {
            'schema': self.raw['schema'],
            'time_reference': self.raw['time_reference'],
            'environment': self.raw['environment'],  # shallow copy!
            'timelines': {},
            'streams': {},
            'events': []
        }

        # Streams first, then events.
        for stream_name, stream in self.raw['streams'].items():

            # Find indices from the original timeline
            tl_name = stream['timeline']
            tl = self.get_timeline(tl_name)  # Raises if not found
            first_i = bisect_left(tl, rel_start_time)  # first in range
            if rel_end_time is None:
                end_i = None
            else:
                end_i = bisect_left(tl, rel_end_time)  # first after range
            # https://docs.python.org/3/library/bisect.html

            # Slice timeline and streams
            # (TODO ensure that timeline slice is done only once)
            if end_i is None:
                sub_timeline = tl[first_i:]
                sub_values = stream['values'][first_i:]
                if 'confidence' in stream:
                    sub_confidence = stream['confidence'][first_i:]
            else:
                sub_timeline = tl[first_i:end_i]
                sub_values = stream['values'][first_i:end_i]  # incl-excl
                if 'confidence' in stream:
                    sub_confidence = stream['confidence'][first_i:end_i]

            # Include stream to the slice if not empty.
            if len(sub_values) > 0:
                slice_raw['timelines'][tl_name] = sub_timeline

                substream = {}
                substream['timeline'] = tl_name
                substream['values'] = sub_values
                # Include confidencies only if they exist beforehand
                if 'confidence' in stream:
                    substream['confidence'] = sub_confidence
                slice_raw['streams'][stream_name] = substream

        # Result:
        #   only needed timelines are included
        #   timelines, values, and confidencies are cut
        #   empty timelines are not included

        def copy_event(event):
            '''Return copy. Extra is shallow copied.'''
            ev = {
                'tags': list(event['tags']),
                'range': list(event['range'])
            }
            if 'extra' in event:
                ev['extra'] = event['extra']
            return ev

        # Then events.
        # Slice events that overlap with the time range.
        # Keep events that are inside.
        # Remove events that are outside.
        for event in self.raw['events']:
            ev_start = event['range'][0]
            ev_end = event['range'][1]

            # # A false approach:
            # # Force into the range
            # upper_limit = rel_end_time
            # lower_limit = rel_start_time
            # f_ev_start = max(min(ev_start, upper_limit), lower_limit)
            # f_ev_end = max(min(ev_end, upper_limit), lower_limit)
            # # If event duration 0, it is outside.
            # # This would accidentally remove purposely 0 duration events.
            # # Testing if 0 duration event would be at begin or end
            # # accidentally removes events that needs to be there.

            if rel_end_time is None:
                if rel_start_time < ev_end:
                    # Event overlaps with the slice
                    if ev_start < rel_start_time:
                        # Event only partially overlaps. Copy & modify.
                        nev = copy_event(event)
                        nev['range'][0] = rel_start_time
                    else:
                        # Event inside the slice. Add it.
                        nev = event
                    slice_raw['events'].append(nev)
            else:
                if ev_start < rel_end_time and rel_start_time < ev_end:
                    # Event overlaps with the slice
                    upper_limit = rel_end_time
                    lower_limit = rel_start_time
                    # Limit event range
                    nev_start = max(min(ev_start, upper_limit), lower_limit)
                    nev_end = max(min(ev_end, upper_limit), lower_limit)
                    nev = copy_event(event)
                    nev['range'] = [nev_start, nev_end]
                    # Add the event.
                    slice_raw['events'].append(nev)

        return CommonV1(slice_raw)

    def slice_by_unix_time(self, start_time, end_time=None):
        '''
        Parameters:
            start_time: unix time in microseconds
            end_time: unix time in microseconds

        Return new CommonV1 object with data only from the time range.
        '''

        # Ensure correct order
        self.assert_range_order(start_time, end_time,
                                'start_time', 'end_time')

        r_start = self.convert_to_relative_time(start_time)
        r_end = self.convert_to_relative_time(end_time)

        return self.slice_by_relative_time(r_start, r_end)

    def slice_by_timeline(self, timeline_name, start_index, end_index=None):
        '''
        Return new CommonV1 object with data only in the time range specified
        by the indices of the timeline.

        Parameters:
            timeline_name: string
            start_index: inclusive
            end_index:

                Optional, exclusive, first element to not be included.
                If None given, slice to the end.

        '''

        # Ensure correct order
        self.assert_range_order(start_index, end_index,
                                'start_index', 'end_index')

        timeline = self.get_timeline(timeline_name)

        # Limit indices
        last_index = len(timeline) - 1
        start_index = min(last_index, max(0, start_index))
        if end_index is not None:
            if end_index > last_index:
                end_index = None
            else:
                # Limit to zero or positive
                end_index = max(0, end_index)

        rel_start_time = timeline[start_index]
        if end_index is None:
            rel_end_time = None
        else:
            rel_end_time = timeline[end_index]

        return self.slice_by_relative_time(rel_start_time, rel_end_time)

    def slice_by_tag(self, tag, index=0):
        '''
        Parameters:
            tag
            index

        Return:
            CommonV1 instance
                that has data only during the tagged event.

        Raise MissingTagException if no events with given tag found.
        Raise IndexError if tag found but index out of range.
        '''
        ev = self.get_event_by_tag(tag, index)
        range_start = ev['range'][0]
        range_end = ev['range'][1]
        return self.slice_by_relative_time(range_start, range_end)

    def slice_first_microseconds(self, n):
        '''
        Parameters:
            n: <integer> microseconds

        Return:
            CommonV1 instance
                that has data only from the first n microseconds, counted
                from the beginning of the slice.
        '''
        range_start = self.get_relative_start_time()
        range_end = range_start + n
        return self.slice_by_relative_time(range_start, range_end)

    # Mutators

    def add_environment(self, env_name, env_value):
        '''
        Add new environment.

        Parameters:
            env_name
                A string, the name of the environment. Use of semantic names
                is recommended.
            env_value
                Any type of JSON-compatible object.
        '''
        # TODO raise error if invalid
        self.raw['environment'][env_name] = env_value

    def add_event(self, tags, start_time, end_time, derived=None, extra=None):
        '''
        Add new event.

        Parameters:
            tags
                A list of tag strings.
            start_time
                A relative time in microseconds. The starting time of event.
            end_time
                A relative time in microseconds. The ending time of event.
            derived
                Optional name of function that was used to derive the event.
            extra
                An optional place for extra info. Typically a dict.
        '''
        if not is_list_of_strings(tags):
            raise CommonV1.InvalidEventException('Invalid tag list')
        if (not is_integer(start_time)) or (not is_integer(end_time)):
            rang = '[' + str(start_time) + ', ' + str(end_time) + ']'
            raise CommonV1.InvalidEventException('Invalid range: ' + rang)
        if start_time > end_time:
            rang = '[' + str(start_time) + ', ' + str(end_time) + ']'
            msg = 'Invalid range order: ' + rang
            raise CommonV1.InvalidEventException(msg)
        new_event = {
            'tags': tags,
            'range': [start_time, end_time],
        }
        if derived is not None:
            if not is_string(derived):
                msg = 'Derived must be a string: ' + str(derived)
                raise CommonV1.InvalidEventException(msg)
            new_event['derived'] = derived
        if extra is not None:
            new_event['extra'] = extra

        self.raw['events'].append(new_event)

    def add_stream(self, stream_name, timeline_name, values, confidence=None):
        '''
        Add a new sequence of sampled data.

        Parameters:
            stream_name
                A string. It is advisable to use semantic names.
            timeline_name
                The name of the timeline that contains the sampling times of
                the values.
            values
                An iterable of values. The length cannot be larger than
                the length of the timeline.
            confidence
                An optional iterable of confidence values. The confidencies
                must be inclusively within 0.0 and 1.0.
        '''
        if timeline_name not in self.raw['timelines']:
            msg = 'Timeline ' + timeline_name + ' not found.'
            raise CommonV1.MissingTimelineException(msg)

        if type(stream_name) is not str:
            raise CommonV1.InvalidStreamException()

        if not hasattr(values, '__iter__'):
            msg = 'Values must be an iterable.'
            raise CommonV1.InvalidStreamException(msg)
        # Convert to list so length can be checked.
        # The specification also requires the internal representation
        # to be a list.
        values = list(values)

        if len(values) != len(self.raw['timelines'][timeline_name]):
            msg = 'Stream and timeline must have equal length.'
            raise CommonV1.InvalidStreamException(msg)

        if confidence is not None:
            if not hasattr(confidence, '__iter__'):
                msg = 'Confidence must be an iterable or None'
                raise CommonV1.InvalidStreamException(msg)
            # Convert to list so length can be checked.
            # The specification also requires the internal representation
            # to be a list.
            confidence = list(confidence)
            if len(confidence) != len(values):
                msg = 'Confidence length must equal to values'
                raise CommonV1.InvalidStreamException(msg)
            # Ensure values inclusively between 0.0 and 1.0
            if any(map(lambda c: c < 0.0 or c > 1.0, confidence)):
                msg = 'Confidence values must be within [0.0, 1.0]'
                raise CommonV1.InvalidStreamException(msg)

        new_stream = {
            'timeline': timeline_name,
            'values': values,
        }
        if confidence is not None:
            new_stream['confidence'] = confidence

        self.raw['streams'][stream_name] = new_stream

    def add_timeline(self, timeline_name, timeline_values):
        '''
        Add a new timeline.

        Parameters:
            timeline_name
                A string that will be referenced from the streams.
            timeline_values
                An iterable of integer microseconds. Will be converted to list.
        '''
        if type(timeline_name) is str and hasattr(timeline_values, '__iter__'):
            self.raw['timelines'][timeline_name] = list(timeline_values)
        else:
            raise CommonV1.InvalidTimelineException()

    def set_time_reference(self, microseconds_from_epoch):
        '''
        Can be used to anonymize data.

        Parameter:
            microseconds_from_epoch, integer.

        Raise:
            InvalidTimeException if not integer.
        '''
        if isinstance(microseconds_from_epoch, int):
            # Note: what if long
            self.raw['time_reference'] = microseconds_from_epoch
        else:
            raise CommonV1.InvalidTimeException('Must be integer.')

    # IO

    def save_events_as_csv(self, tags, target_file_path, delimiter='\t'):
        '''
        Save CSV file with each event as row and with columns:
            start_time: event start time in unix time
            end_time: event end time in unix time
            <1st tag>: 1 = event has tag, 0 = event does not have this tag
            <2nd tag>:
            ...

        The file includes header row.
        '''
        def iterevents():
            for ev in self.iter_events_by_tags(tags):
                row = {
                    'start_time': ev['range'][0],
                    'end_time': ev['range'][1]
                }
                for tag in tags:
                    if tag in ev['tags']:
                        row[tag] = 1
                    else:
                        row[tag] = 0
                yield row

        # Define order of columns
        headers = ['start_time', 'end_time'] + tags
        write_dictlist_as_csv(target_file_path, iterevents(),
                              headers=headers, delimiter=delimiter)

    def save_timeline_as_csv(self, timeline_name, target_file_path,
                             delimiter='\t'):
        '''
        Store timeline and its associated streams in a comma separated values
        format. Note that neither events nor environments are included.
        '''
        # Find timeline
        tl = self.get_timeline(timeline_name)

        # Build the headers in practical order.
        headers = ['gazelib/time/microseconds']

        # Find streams on the timeline
        streams = {}
        for stream_name, stream in self.raw['streams'].items():
            if stream['timeline'] == timeline_name:
                streams[stream_name] = stream['values']
                headers.append(stream_name)
                if 'confidence' in stream:
                    conf_name = stream_name + '/confidence'
                    streams[conf_name] = stream['confidence']
                    headers.append(conf_name)

        # Define how to iterate over streams to get rows
        def iterstreams():
            for index in range(len(tl)):
                d = {}
                d['gazelib/time/microseconds'] = tl[index]
                for stream_name, stream in streams.items():
                    d[stream_name] = stream[index]
                yield d

        write_dictlist_as_csv(target_file_path, iterstreams(),
                              headers=headers, delimiter=delimiter)

    def save_as_json(self, target_file_path, human_readable=False):
        '''
        Store the content in gazelib/common/v1 in a JSON file.

        Parameters:
            target_file_path
                Absolut or relative file path as string.
            human_readable
                Set True for new lines and indentation. Default to False.
        '''
        if human_readable:
            write_fancy_json(target_file_path, self.raw)
        else:
            write_json(target_file_path, self.raw)



# -*- coding: utf-8 -*-


class ExtrapolationError(Exception):
    pass


def fill_gaps(l):
    '''
    Given a list of values, extrapolate None values by the last
    known non-None value. If no preceding non-None values were found, copy
    the first known value to those.

    Paremeters:
        l: list of values

    Throw
        ExtrapolationError
            if list does not contains any non-null values

    Return
        list without gaps
    '''
    nl = []  # new, filled list

    if len(l) < 1:
        raise ExtrapolationError('Empty list cannot be extrapolated')

    first_nonnull = None

    # Find first non-nulls
    for p in l:
        if p is not None:
            first_nonnull = p
            break

    if first_nonnull is None:
        # No good values found
        msg = 'No non-null values to fill with: ' + str(first_nonnull)
        raise ExtrapolationError(msg)

    prev_nonnull = first_nonnull

    for p in l:
        if p is None:
            nl.append(prev_nonnull)
        else:
            prev_nonnull = p
            nl.append(p)

    return nl



# -*- coding: utf-8 -*-


def isNotNone(x):
    return x is not None


def isNotNone2d(x, y):
    return x is not None and y is not None


def get_valid_sublists(l, validator=isNotNone):
    '''
    Return list of lists, where each sublist contains only valid values.
    In other words, the given list is splitted by its invalid values.

    Parameters:
        l:
            A list
        validator:
            A boolean function that returns true if element is valid.
            Defaults to testing if value is not None.

    Example::

        >>> get_valid_sublists([1, 2, None, 3])
        [[1, 2], [3]]

    '''
    sublists = []
    cur_sublist = []

    for x in l:
        if validator(x):
            cur_sublist.append(x)
        else:
            if len(cur_sublist) > 0:
                # End current sublist and start a new one.
                sublists.append(cur_sublist)
                cur_sublist = []

    # If last value was valid, one sublist remains to be added.
    if len(cur_sublist) > 0:
        sublists.append(cur_sublist)

    return sublists


def get_valid_sublists_2d(xs, ys, validator=isNotNone2d):
    '''
    Return list of lists

    Example result::

        [
            ([1,2,3,...], [0,3,5,...]),
            ([4,6,2,...], [3,0,3,...]),
            ...
        ]

    '''
    sublists = []
    cur_sublist_x = []
    cur_sublist_y = []

    for x, y in zip(xs, ys):
        if validator(x, y):
            cur_sublist_x.append(x)
            cur_sublist_y.append(y)
        else:
            if len(cur_sublist_x) > 0:
                # End current sublist and start a new one
                sublists.append((cur_sublist_x, cur_sublist_y))
                cur_sublist_x = []
                cur_sublist_y = []

    # If last value was valid, one sublist remains to be added.
    if len(cur_sublist_x) > 0:
        sublists.append((cur_sublist_x, cur_sublist_y))

    return sublists



# -*- coding: utf-8 -*-
'''
Common statistical tools that do not depend on CommonV1.
'''


def maximum(l):
    '''Maximum of a list. Allows None values. Return None if no valid value.'''
    m = None
    for item in l:
        if item is not None:
            if m is not None:
                if item > m:
                    m = item
            else:
                m = item
    return m


def minimum(l):
    '''Minimum of a list. Allows None values. Return None if no valid value.'''
    m = None
    for item in l:
        if item is not None:
            if m is not None:
                if item < m:
                    m = item
            else:
                m = item
    return m


def arithmetic_mean(l):
    '''Return arithmetic mean of list. Return None if empty list.
    Allows None values and does not take them into consideration.'''
    # Based loosely on http://stackoverflow.com/a/7716358/638546
    summ = 0.0
    count = 0
    for item in l:
        if item is not None:
            summ += item
            count += 1
    return summ / count if count > 0 else None


def weighted_arithmetic_mean(l, w):
    '''
    Parameters:
        l: a list of numerical values. Nones are allowed.
        w: a list of numerical weights. Nones are regarded as zeros.

    Return
        float. None if no non-none weights or values were given.
    '''
    suml = 0.0
    sumw = 0.0
    for item, weight in zip(l, w):
        if item is not None:
            if weight is not None:
                suml += item * weight
                sumw += weight
    return suml / sumw if sumw > 0.0 else None


def deltas(l):
    '''
    Return list of differences between consecutive list items.
    The length of the returned list is one smaller than the given list.
    For an increasing sequence, the deltas are positive.
    For a decreasing sequence, the deltas are negative.
    '''
    if len(l) == 0 or len(l) == 1:
        return []
    diffs = []
    for i in range(len(l) - 1):
        diffs.append(l[i + 1] - l[i])
    return diffs



# -*- coding: utf-8 -*-
'''
Find a linear saccade from the data.
'''
from gazelib.containers import CommonV1
from gazelib.preprocessing import fill_gaps, ExtrapolationError
import scipy.signal
import saccademodel


def fit(g):
    '''
    Parameter:
        g: Gaze data as CommonV1 object

    Require streams:
        gazelib/gaze/left_eye_x_relative
        gazelib/gaze/left_eye_y_relative
        gazelib/gaze/right_eye_x_relative
        gazelib/gaze/right_eye_y_relative

    Raise:
        InsufficientDataException: if streams are missing or they are empty.

    Return::

        {
            'type': 'gazelib/gaze/saccade',
            'start_time_relative': <int microseconds>
            'end_time_relative': <int microseconds>
            'mean_squared_error': <float>
        }

    '''
    g.assert_has_streams([
        'gazelib/gaze/left_eye_x_relative',
        'gazelib/gaze/left_eye_y_relative',
        'gazelib/gaze/right_eye_x_relative',
        'gazelib/gaze/right_eye_y_relative'
    ])
    # Timeline names
    l_tl_name = g.get_stream_timeline_name('gazelib/gaze/left_eye_x_relative')
    r_tl_name = g.get_stream_timeline_name('gazelib/gaze/right_eye_x_relative')

    lx = g.raw['streams']['gazelib/gaze/left_eye_x_relative']['values']
    ly = g.raw['streams']['gazelib/gaze/left_eye_y_relative']['values']
    rx = g.raw['streams']['gazelib/gaze/right_eye_x_relative']['values']
    ry = g.raw['streams']['gazelib/gaze/right_eye_y_relative']['values']

    # Forward fill
    try:
        lx_fill = fill_gaps(lx)
        ly_fill = fill_gaps(ly)
        rx_fill = fill_gaps(rx)
        ry_fill = fill_gaps(ry)
    except ExtrapolationError:
        # Only nones or empty
        msg = 'Cannot find saccade from empty data.'
        raise CommonV1.InsufficientDataException(msg)

    # Median filter
    # Required to remove non-Gaussian noise i.e. random outliers
    # Saccademodel handles Gaussian noise.
    lx_filt = scipy.signal.medfilt(lx_fill, 5)
    ly_filt = scipy.signal.medfilt(ly_fill, 5)
    rx_filt = scipy.signal.medfilt(rx_fill, 5)
    ry_filt = scipy.signal.medfilt(ry_fill, 5)

    # Pointlists for saccademodel
    lpl = [[x, y] for x, y in zip(lx_filt, ly_filt)]
    rpl = [[x, y] for x, y in zip(rx_filt, ry_filt)]

    # Results
    try:
        lresults = saccademodel.fit(lpl)
        rresults = saccademodel.fit(rpl)
    except saccademodel.interpolate.InterpolationError:
        msg = 'Cannot find saccade from empty data.'
        raise CommonV1.InsufficientDataException(msg)

    # Pick one with smallest error
    lerr = lresults['mean_squared_error']
    rerr = rresults['mean_squared_error']
    if lerr < rerr:
        results = lresults
        tl_name = l_tl_name
    else:
        results = rresults
        tl_name = r_tl_name

    # Convert measured saccade end and start to times.
    lensource = len(results['source_points'])
    lensaccade = len(results['saccade_points'])

    start_index = max(0, lensource - 1)  # do not let below zero
    end_index = max(0, lensource + lensaccade - 1)  # do not let above length

    start = g.get_relative_time_by_index(tl_name, start_index)
    end = g.get_relative_time_by_index(tl_name, end_index)

    return {
        'type': 'gazelib/gaze/saccade',
        'start_time_relative': start,  # microseconds
        'end_time_relative': end,
        'mean_squared_error': results['mean_squared_error']
    }



# -*- coding: utf-8 -*-
import bokeh.plotting as plotting
from . import utils
import gazelib.preprocessing as gpre

# For printing environments
import yaml

# Custom HTML templates are required by custom HTML such as tables.
# The following setup is needed for custom HTML with bokeh.
# See http://bokeh.pydata.org/en/0.10.0/docs/user_guide/embed.html
# See http://bokeh.pydata.org/en/latest/docs/reference/embed.html
# For FileSystemLoader, see https://gist.github.com/wrunk/1317933
from bokeh.resources import CDN
from bokeh.embed import file_html
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound
import os


def render_path(common, output_html_filepath, title='Path'):
    '''
    Create a HTML-based visualization of the gaze path.
    '''
    lxs = common.get_stream_values('gazelib/gaze/left_eye_x_relative')
    lys = common.get_stream_values('gazelib/gaze/left_eye_y_relative')
    rxs = common.get_stream_values('gazelib/gaze/right_eye_x_relative')
    rys = common.get_stream_values('gazelib/gaze/right_eye_y_relative')

    p = plotting.figure(title=title, x_axis_label='X', y_axis_label='Y',
                        plot_width=640, plot_height=480,
                        x_range=(-0.2, 1.2), y_range=(-0.2, 1.2))

    # Render screen rectangle
    p.quad(left=[0.0], right=[1.0], top=[0.0], bottom=[1.0],
           line_width=1, line_color='black', fill_alpha=0)

    # Visualize only valid points.
    # We do that by visualizing valid subpaths

    def validator(x):
        return x[0] is not None and x[1] is not None

    lpaths = utils.get_valid_sublists(zip(lxs, lys), validator)
    rpaths = utils.get_valid_sublists(zip(rxs, rys), validator)

    for valid_path in lpaths:
        valid_xs = list(map(lambda x: x[0], valid_path))
        valid_ys = list(map(lambda x: x[1], valid_path))

        p.cross(valid_xs, valid_ys, size=10, line_color='blue')
        p.line(valid_xs, valid_ys, line_width=1, line_color='blue')

    for valid_path in rpaths:
        valid_xs = list(map(lambda x: x[0], valid_path))
        valid_ys = list(map(lambda x: x[1], valid_path))

        p.cross(valid_xs, valid_ys, size=10, line_color='red')
        p.line(valid_xs, valid_ys, line_width=1, line_color='red')
        # Red as Right

    # Display start points.
    lfirst = lpaths[0][0]
    rfirst = rpaths[0][0]
    # Width and height in x_axis and y_axis units.
    p.oval(x=lfirst[0], y=lfirst[1], width=0.17, height=0.2,
           line_color='blue', fill_alpha=0)
    p.oval(x=rfirst[0], y=rfirst[1], width=0.17, height=0.2,
           line_color='red', fill_alpha=0)

    plotting.output_file(output_html_filepath, title)
    plotting.save(p)


def render_path_for_each_event(common, event_tag, output_html_filepath):
    pass


def render_overview(common, output_html_filepath, title='Overview',
                    emphasize_gaps=False):
    '''
    Create HTML-based visualization from all streams and events.
    Does not understand stream or event semantics like gaze paths for example.

    Can be slow if streams contain large number of gaps.

    Parameters:
        common: A CommonV1 object
        output_html_filepath: a filepath as string
        title: HTML page title as string
        emphasize_gaps: if False, do not show red gaps. Makes it quicker.
    '''

    # Collect figures together
    figs = []

    def to_ms(micros):
        return int(round(micros / 1000))

    def pick_color(d):
        '''
        We visually discriminate original and derived data.

        Parameters:
          d: event or stream object.

        Return color string
        '''
        if 'derived' in d:
            return 'blue'
        return 'black'

    #########
    # Environment constants
    #########

    # Build dict
    env_names = common.list_environment_names()
    envs = {name: common.get_environment(name) for name in env_names}
    # Make human readable
    # yaml.safe_dump instad of yaml.dump to avoid annoying unicode tags
    # See http://stackoverflow.com/a/1950399/638546
    env_yaml = yaml.safe_dump(envs, default_flow_style=False)
    env_html = '<pre>' + env_yaml + '</pre>'

    #########
    # Streams
    #########

    stream_names = sorted(common.list_stream_names())
    for stream_name in stream_names:
        stream = common.get_stream(stream_name)
        x = list(map(to_ms, common.get_timeline(stream['timeline'])))
        y = stream['values']
        color = pick_color(stream)

        fig = plotting.figure(title=stream_name, x_axis_label='time (ms)',
                              plot_width=1000, plot_height=300,
                              toolbar_location=None)

        if emphasize_gaps:
            # Emphasize gaps with red. Slow if data very gapped.
            # Get valid substreams
            sl = utils.get_valid_sublists_2d(x, y)

            for xs, ys in sl:
                fig.line(xs, ys, line_color=color)

            # Emphasize gaps with red line.
            # Loop pairwise, fill gaps.
            # TODO Optimize. Currently very slow.
            for i in range(len(sl) - 1):
                xs0, ys0 = sl[i]
                xs1, ys1 = sl[i + 1]
                # Extrapolate from last known value
                x0 = xs0[-1]
                x1 = xs1[0]
                y = ys0[-1]
                fig.line(x=[x0, x1], y=[y, y], line_width=1, line_color='red')
        else:
            # Fill gaps and draw single line.
            # This should be much faster.
            yy = gpre.fill_gaps(y)
            fig.line(x=x, y=yy, line_width=1, line_color=color)

        figs.append(fig)

    ########
    # Events
    ########

    # Create a row for each event. X is time.
    evs = common.list_events()

    # Visualize in start time, then duration order.
    # Recent topmost, hence minus. If start time same, longest topmost.
    def order_key(ev):
        t0 = -ev['range'][0]
        dur = ev['range'][1] + t0
        return (t0, dur)

    evs = sorted(evs, key=order_key)

    # Plot height: it is dependent on number of events.
    # With one event, we want it still be visible, thus the constant.
    plot_height = (100 + 50 * len(evs))
    fig = plotting.figure(title='Events', y_range=(-1, len(evs)),
                          plot_width=1000,
                          plot_height=plot_height,
                          x_axis_label='time (ms)',
                          toolbar_location=None)

    # Hide event indices and draw custom text instead.
    fig.yaxis.visible = None

    def get_text_position(x0, x1, minx, maxx):
        '''Change text position based on event's horizontal position so that
        the rightmost text do not go over the right edge.'''
        # Event mean, normalized to range
        c = (x0 + x1) / 2 - minx
        # Max range
        r = maxx - minx

        if c < 2 * r / 3:
            return (x0, 'left')
        # Center align was not good because text formed
        # visually annoying justification.
        # if c < 2 * r / 3:
        #     return ((x0 + x1) / 2, 'center')
        return (x1, 'right')

    # Limit for horizontal coordinates. Helps to position text.
    evs_minx = to_ms(evs[-1]['range'][0])
    evs_maxx = to_ms(evs[0]['range'][1])

    for i, ev in enumerate(evs):
        t0 = to_ms(ev['range'][0])
        t1 = to_ms(ev['range'][1])

        color = pick_color(ev)

        x, align = get_text_position(t0, t1, evs_minx, evs_maxx)
        fig.text(text=[', '.join(ev['tags'])], y=[i + 0.1], x=[x],
                 text_align=align, text_color=color)

        fig.line(x=[t0, t1], y=[i, i], line_width=6, line_color=color)
        # Mark where events start and end.
        fig.line(x=[t0, t0], y=[i - 0.1, i + 0.1],
                 line_width=2, line_color=color)
        fig.line(x=[t1, t1], y=[i - 0.1, i + 0.1],
                 line_width=2, line_color=color)

    figs.append(fig)

    # Make human readable
    evs_rev = list(reversed(evs))
    evs_yaml = yaml.safe_dump(evs_rev, default_flow_style=False)
    evs_html = '<pre>' + evs_yaml + '</pre>'

    ###########
    # Render HTML
    ##########

    this_dir = os.path.dirname(os.path.abspath(__file__))
    template_dir = os.path.join(this_dir, 'templates')
    jinja2loader = FileSystemLoader(template_dir)
    jinja2env = Environment(loader=jinja2loader)
    try:
        overview_template = jinja2env.get_template('overview.html')
    except TemplateNotFound as ex:
        print('Tried template dir: ' + template_dir)
        raise ex

    html = file_html(figs, CDN, title=title,
                     template=overview_template,
                     template_variables={'environment': env_html,
                                         'events': evs_html})
    # Save as html file.
    with open(output_html_filepath, 'w') as f:
        f.write(html)



# -*- coding: utf-8 -*-
'''
Gazelib, a toolkit for gaze analysis.
'''
from gazelib import package
__version__ = package.version  # noqa

from gazelib import containers  # noqa
from gazelib import io  # noqa
from gazelib import legacy  # noqa
# import gazelib.settings  # noqa
from gazelib import statistics  # noqa
# import gazelib.validation  # noqa
# from . import conversion  # Must be imported separately
